import os
import sys
import shutil
import numpy as np
import matplotlib.pyplot as plt

path = sys.argv[1]

path = os.path.abspath(path)
FIELDS = os.path.join(path, "fields")
FRAMES = os.path.join(path, "frames")
MOVIE = os.path.join(path, "sim.mp4")
shutil.rmtree(FRAMES, ignore_errors=True)
os.makedirs(FRAMES, exist_ok=True)
shutil.rmtree(MOVIE, ignore_errors=True)

g = np.load(os.path.join(path, "temp", 'g.npy')).T
gmax = np.max(np.abs(g))

dir = os.path.join(path, "temp", 'fields')
umax = 0
fns = sorted(os.listdir(dir), key=lambda x: float(x[0:-4]))
for fn in fns:
    a = np.load(os.path.join(dir, fn))
    v = np.max(np.abs(a))
    if umax < v:
        umax = v

for fn in fns:
    name = fn[0:-4]
    a = np.load(os.path.join(FIELDS, fn)).T
    fig, axs = plt.subplots(1, 2)
    axs[1].imshow(a, cmap='seismic', origin='lower',
                  vmin=-umax, vmax=umax)
    axs[0].imshow(-g, cmap='gray',
                  origin='lower', vmin=-gmax, vmax=0)
    plt.savefig(os.path.join(FRAMES, f"{name}.png"))
    plt.close(fig)



import math
from IPython.display import Image
from PIL import Image as PILImage
import os

# import pymeshfix
import json
from statistics import median
from .constants import *
import gdsfactory as gf
from gdsfactory.cross_section import Section
from .constants import *
from .materials import *
import pyvista as pv

try:
    from IPython.display import display
except ImportError:
    pass
# import pyvista as pv

# from .picToGDS import main

from math import ceil, cos, pi, sin, tan
import matplotlib.pyplot as plt
import numpy as np
from gdsfactory.generic_tech import LAYER_STACK, LAYER
import copy
import shutil
import trimesh

# from gdsfactory import LAYER_VIEWS

tol = 0.001


def get(c, i):
    try:
        return c[i]
    except:
        try:
            return getattr(c, i)
        except:
            return None


def arange(a, b, d):
    ret = np.linspace(a, b, round((b - a) / (d)) + 1).tolist()
    return ret


def trim(x, dx):
    return round(x / dx) * dx


def extend(endpoints, wm):
    v = endpoints[1] - endpoints[0]
    v = v / np.linalg.norm(v)
    return [(endpoints[0] - wm * v).tolist(), (endpoints[1] + wm * v).tolist()]


def portsides(c):
    ports = c.ports
    bbox = c.bbox_np()
    res = [[], [], [], []]
    xmin0, ymin0 = bbox[0]
    xmax0, ymax0 = bbox[1]
    for p in ports:
        x, y = np.array(p.center) / 1e0

        if abs(x - xmin0) < tol:
            res[2].append(p.name)
        if abs(x - xmax0) < tol:
            res[0].append(p.name)
        if abs(y - ymin0) < tol:
            res[3].append(p.name)
        if abs(y - ymax0) < tol:
            res[1].append(p.name)
    return res


def add_bbox(c, layer, nonport_margin=0):  # , dx=None):
    bbox = c.bbox_np()
    xmin0, ymin0 = bbox[0]
    xmax0, ymax0 = bbox[1]
    l = xmax0 - xmin0
    w = ymax0 - ymin0

    # l = dx*np.ceil((xmax0-xmin0)/dx)
    # w = dx*np.ceil((ymax0-ymin0)/dx)

    # if dx is not None:
    #     if nonport_margin is None:
    #         nonport_margin = dx
    # if nonport_margin is None:
    #     nonport_margin = 0
    margin = nonport_margin
    xmin, ymin, xmax, ymax = (
        xmin0 - margin,
        ymin0 - margin,
        xmin0 + l + margin,
        ymin0 + w + margin,
    )

    for p in c.ports:
        # p = c.ports[k]
        x, y = np.array(p.center) / 1e0
        if abs(x - xmin0) < tol:
            xmin = x
        if abs(x - xmax0) < tol:
            xmax = x
        if abs(y - ymin0) < tol:
            ymin = y
        if abs(y - ymax0) < tol:
            ymax = y
    p = [(xmin, ymin), (xmax, ymin), (xmax, ymax), (xmin, ymax)]
    _c = gf.Component()
    _c << c

    if type(layer[0]) is int:
        layer = [layer]
    for layer in layer:
        layer = tuple(layer)
        _c.add_polygon(p, layer=layer)
    for port in c.ports:
        _c.add_port(name=port.name, port=port)
    return _c

    # def pic2gds(fileName, sizeOfTheCell, layerNum=1, isDither=False, scale=1):
    main(fileName, sizeOfTheCell, layerNum, isDither, scale)
    return "image.bmp", "image.gds"


def finish(c, name):
    c.add_label(name, position=c.bbox_np()[1])


def normal_from_orientation(orientation):
    return [cos(orientation / 180 * pi), sin(orientation / 180 * pi)]


def generate_background_mesh(bounds, resolution=20, eps=1e-6):
    x_min, x_max, y_min, y_max, z_min, z_max = bounds
    grid_x, grid_y, grid_z = np.meshgrid(
        np.linspace(x_min - eps, x_max + eps, resolution),
        np.linspace(y_min - eps, y_max + eps, resolution),
        np.linspace(z_min - eps, z_max + eps, resolution),
        indexing="ij",
    )
    return pv.StructuredGrid(grid_x, grid_y, grid_z).triangulate()


def sizing_function(
    points, focus_point=np.array([0, 0, 0]), max_size=1.0, min_size=0.1
):
    distances = np.linalg.norm(points - focus_point, axis=1)
    return np.clip(max_size - distances, min_size, max_size)


def material_voxelate(c, zmin, zmax, layers, layer_stack, materials, path):
    shutil.rmtree(path, ignore_errors=True)
    os.makedirs(path, exist_ok=True)
    stacks = sum(
        [
            [
                [v.mesh_order, v.material, tuple(layer), k, v.info]
                for k, v in get_layers(layer_stack, layer, withkey=True)
            ]
            for layer in layers
        ],
        [],
    )
    c.flatten()
    stacks = sorted(stacks, key=lambda x: x[0])
    layer_stack_info = dict()
    # raise NotImplementedError("This is a stub")
    lb, ub = c.bbox_np()
    # bbox = [[**lb, zmin], [**ub, zmax]]
    bbox = [[lb[0], lb[1], zmin], [ub[0], ub[1], zmax]]
    layers = [x[2] for x in stacks]

    layer_views = copy.deepcopy(LAYER_VIEWS)

    i = 1
    for stack in stacks:
        order = stack[0]
        m = stack[1]
        l1, l2 = layer = stack[2]
        k = stack[3]

        layer_views.layer_views[k] = LayerView(layer=layer, color="blue", visible=True)
        _layer_stack = copy.deepcopy(layer_stack)
        get(_layer_stack, "layers").clear()

        d = copy.deepcopy(get(layer_stack, "layers")[k])
        if d.zmin <= zmax and d.bounds[1] >= zmin:
            get(_layer_stack, "layers")[k] = d
            # _d.bounds = (_d.zmin, _d.zmin+_d.thickness)
            # origin = (c.extract([layer]).bbox_np()-c.bbox_np())[0].tolist()

            meshes = c.to_3d(
                layer_stack=_layer_stack,
                layer_views=layer_views,
                exclude_layers=set(layers) - {layer},
            ).dump()
            for mesh in meshes:
                OBJ = os.path.join(path, f"{i}_{m}_unnamed{i}1.obj")
                i += 1
                # print(mesh.bounds)
                mesh = mesh.slice_plane([0, 0, zmin], [0, 0, 1])
                mesh = mesh.slice_plane([0, 0, zmax], [0, 0, -1])
                # print(mesh.bounds)
                trimesh.exchange.export.export_mesh(mesh, OBJ, "obj")
                # mesh = pv.get_reader(OBJ).read()
                # # tetgen.refine_with_background_mesh(pv.Triangle(), mesh).save(OBJ)

                # OBJ = OBJ.replace(".stl", ".obj")
                # bg_mesh = generate_background_mesh(mesh.bounds)
                # bg_mesh.point_data["target_size"] = sizing_function(bg_mesh.points)

                # tet = tetgen.TetGen(mesh)
                # tet.make_manifold()
                # tet.tetrahedralize(
                #     # nobisect=True,
                #     # quality=True,
                #     # minratio=1.1,
                #     # mindihedral=10,
                #     bgmesh=bg_mesh,
                #     verbose=True,
                # )
                # tet.mesh.save(OBJ)

            # pymeshfix.clean_from_file(OBJ, OBJ)
            # repair_obj_open3d(OBJ, OBJ)

            # STL = os.path.abspath(os.path.join(path, f"{k}.stl"))
            # gf.export.to_stl(c, STL, layer_stack=_layer_stack)

            # STL = os.path.join(path, f'{k}_{l1}_{l2}.stl')
            # pymeshfix.clean_from_file(STL, STL)
            # STL1 = os.path.join(path, f'{order}_{m}_{k}.stl')
            # os.replace(STL, STL1)

            # mesh = pv.read(STL)
            # im = stl_to_array(mesh, dl*unit, bbox)
            # np.save(os.path.join(path, f'{k}.npy'), im)

            layer_stack_info[k] = {
                "layer": (l1, l2),
                "zmin": d.zmin,
                "thickness": d.thickness,
                # "material": matname(m),
                "mesh_order": stack[0],
                # "origin": origin,
            }
    return layer_stack_info


def get_layers(layer_stack, layer, withkey=False):
    r = []
    d = get(layer_stack, "layers").items()

    for k, x in d:
        l = get(x, "layer")
        if l is not None:
            t = get(l, "layer")
            if t is not None and tuple(t) == tuple(layer):
                if withkey:
                    x = k, x
                r.append(x)
    if r:
        return r

    for k, x in d:
        l = get(x, "derived_layer")
        if l is not None:
            t = get(l, "layer")
            if t is not None and tuple(t) == tuple(layer):
                if withkey:
                    x = k, x
                r.append(x)
    return r


def wavelength_range(center, bandwidth, length=3):
    f1 = 1 / (center + bandwidth / 2)
    f2 = 1 / (center - bandwidth / 2)
    hw = (f2 - f1) / 2
    f1 = 1 / center - hw
    f2 = 1 / center + hw
    return sorted([1 / x for x in np.linspace(f1, f2, length).tolist()])


def wrap(wavelengths):
    if isinstance(wavelengths, float) or isinstance(wavelengths, int):
        wavelengths = [[wavelengths]]
    elif isinstance(wavelengths[0], float) or isinstance(wavelengths[0], int):
        wavelengths = [wavelengths]
    return wavelengths


def save_problem(prob, path):
    path = os.path.abspath(path)
    bson_data = json.dumps(prob)
    # prob["component"] = c0

    path = prob["path"]
    if not os.path.exists(path):
        os.makedirs(path)
        #   compiling julia code...
        #   """)
    prob_path = os.path.join(path, "problem.json")
    with open(prob_path, "w") as f:
        # Write the BSON data to the file
        f.write(bson_data)
    print("using simulation folder", path)


def load_prob(path):
    path = os.path.abspath(path)
    print(f"loading problem from {path}")
    return json.loads(open(os.path.join(path, "problem.json"), "rb").read())


def create_gif(image_path, output_path, duration):
    """
    Creates a GIF from a list of image paths.

    Args:
        image_paths: A list of file paths to the images.
        output_path: The output path for the generated GIF.
        duration: The duration of each frame in milliseconds (default: 200).
    """
    image_paths = os.listdir(image_path)
    image_paths = sorted(image_paths, key=lambda x: float(x[0:-4]))
    frames = [PILImage.open(os.path.join(image_path, f)) for f in image_paths]

    # Ensure all frames have the same palette if necessary
    for i in range(len(frames)):
        if frames[i].mode != "RGB":
            frames[i] = frames[i].convert("RGB")

    frames[0].save(
        output_path,
        save_all=True,
        append_images=frames[1:],
        duration=duration,
        loop=0,  # 0 means infinite loop
    )


def make_movie(path):
    # framerate = 10 / load_prob(path)["saveat"]
    framerate = 2
    duration = 1000 / framerate
    FRAMES = os.path.join(path, "frames")
    MOVIE = os.path.join(path, "sim.mp4")
    # shutil.rmtree(MOVIE, ignore_errors=True)
    GIF = os.path.join(path, "sim.gif")
    create_gif(FRAMES, GIF, duration)
    return Image(open(GIF, "rb").read())
    # if saveat is None:
    #     saveat = load_prob(path)["saveat"]

    # g = np.load(os.path.join(path, "temp", 'g.npy')).T
    # gmax = np.max(np.abs(g))

    # dir = os.path.join(path, "temp", 'fields')
    # umax = 0
    # fns = sorted(os.listdir(dir), key=lambda x: float(x[0:-4]))
    # for fn in fns:
    #     a = np.load(os.path.join(dir, fn))
    #     v = np.max(np.abs(a))
    #     if umax < v:
    #         umax = v

    # for fn in fns:
    #     name = fn[0:-4]
    #     a = np.load(os.path.join(FIELDS, fn)).T
    #     fig, axs = plt.subplots(1, 2)
    #     axs[1].imshow(a, cmap='seismic', origin='lower',
    #                   vmin=-umax, vmax=umax)
    #     axs[0].imshow(-g, cmap='gray',
    #                   origin='lower', vmin=-gmax, vmax=0)
    #     plt.savefig(os.path.join(FRAMES, f"{name}.png"))
    #     plt.close(fig)

    # fns = sorted(os.listdir(FRAMES), key=lambda x: float(x[0:-4]))
    # frame = height = width = layers = video = None
    # for i, fn in enumerate(fns):
    #     frame = cv2.imread(os.path.join(FRAMES, fn))
    #     if i == 0:
    #         height, width, layers = frame.shape
    #         video = cv2.VideoWriter(
    #             MOVIE, 0x7634706d, saveat, (width, height))

    #     video.write(frame)

    # cv2.destroyAllWindows()
    # video.release()



import numpy as np
import EMpy
import os
import sys
import json

path = sys.argv[1]
# data = np.load(os.path.join(path, "args.npz"))
data = json.loads(open(os.path.join(path, "args.json"), "rb").read())
eps = data["eps"]
eps = np.array(eps).T
dl = data["dl"]
λ = data["center_wavelength"]
neigs = data["neigs"]
name = data["name"]
is2d = data["is2d"]

m, n = eps.shape
# print(m, n)

m += 1
n += 1
x = np.linspace(0.5*dl, (m-.5)*dl, m)
y = np.linspace(0.5*dl, (n-.5)*dl, n)


def ϵfunc(x_, y_):
    return eps


tol = 1e-6
if is2d:
    print('2d')
    solver = EMpy.modesolvers.FD.VFDModeSolver(
        λ, x, y, ϵfunc,  "AA00").solve(2*neigs, tol)
    modes = solver.modes
else:
    solver = EMpy.modesolvers.FD.VFDModeSolver(
        λ, x, y, ϵfunc,  "0000").solve(neigs, tol)
    modes = solver.modes
    # modes = sorted(solver.modes, key=lambda x: -np.abs(x.neff))


neffs = [np.real(m.neff) for m in modes]
modes = [{k: m.get_field(k, x, y) for k in [
    "Ex", "Ey", "Ez", "Hx", "Hy", "Hz"]} for m in modes]
for i, mode in enumerate(modes):
    np.savez(os.path.join(path, f'{name}_mode_{i}.npz'), **modes[i])



import matplotlib.pyplot as plt
from math import floor
from statistics import mode
from .setup import *
from ..constants import *
from ..layers import *
from ..utils import *
from ..materials import *
import gdsfactory as gf
from copy import deepcopy
import matplotlib.pyplot as plt
import numpy as np

from gdsfactory.cross_section import Section
from gdsfactory.generic_tech import LAYER_STACK, LAYER


def make_sim_problem(
    path,
    wavelengths=[],
    center_wavelength=None,
    frequencies=[],
    center_frequency=None,
    sources=[],
    monitors=[],
    nres=None,
    dx=None,
    layer_stack={},
    materials={},
    study="simulation",
    dtype="float32",
    Ttrans=None,
    Tss=None,
    Tssmin=None,
    wl_res=0.01,
    margins=[[0, 0, 0], [0, 0, 0]],
    df=None,
    gpu=None,
    dorun=True,
):
    if not wavelengths:
        if not center_frequency:
            center_frequency = mode(frequencies)

    wavelengths, center_wavelength, T = adjust_wavelengths(
        wavelengths, center_wavelength, wl_res
    )

    GEOMETRY = os.path.join(path, "geometry")
    bbox = [[None, None, None], [None, None, None]]
    for material in os.listdir(GEOMETRY):
        for fn in os.listdir(os.path.join(GEOMETRY, material)):
            if fn.lower().endswith(".stl"):
                STL = os.path.join(GEOMETRY, material, fn)
                print(STL)
                pymeshfix.clean_from_file(STL, STL)
                mesh = pv.read(STL)
                for i, v, w in zip(
                    range(3), bbox[0], [mesh.bounds[i] for i in [0, 2, 4]]
                ):
                    if v is None or w < v:
                        bbox[0][i] = w
                for i, v, w in zip(
                    range(3), bbox[1], [mesh.bounds[i] for i in [1, 3, 5]]
                ):
                    if v is None or w > v:
                        bbox[1][i] = w
    bbox[0] = (np.array(bbox[0]) - np.array(margins[0])).tolist()
    bbox[1] = (np.array(bbox[1]) + np.array(margins[1])).tolist()

    bbox[1] = [a + dx * floor((b - a) / dx) for (a, b) in zip(bbox[0], bbox[1])]

    print(bbox)

    L = (bbox[1][0] - bbox[0][0], bbox[1][1] - bbox[0][1], bbox[1][2] - bbox[0][2])
    if not Tss:
        Tss = T if len(wavelengths) > 1 else None
    prob = {
        "class": "gen",
        "sources": sources,
        "monitors": monitors,
        "nres": nres,
        "study": study,
        "dtype": dtype,
        "center_wavelength": center_wavelength,
        "layer_stack": layer_stack,
        "materials": materials,
        "L": L,
        "gpu_backend": gpu,
        "Ttrans": Ttrans,
        "Tss": Tss,
        "Tssmin": Tssmin,
        "wavelengths": wavelengths,
        "frequencies": frequencies,
        "df": df,
        "dorun": dorun,
    }
    PROB = os.path.join(path, "problem.json")
    with open(PROB, "w") as f:
        json.dump(prob, f)
    return prob

    # l = [k for k in imow if port_number(k) == pi]
    # if not l:
    #     imow[f"o{pi}@{mi}"] = []
    # else:
    #     k = l[0]
    #     mn = max(mode_number(k), mi)
    #     if mn != mode_number(k):
    #         imow[i] = imow[k]
    #         del imow[k]

    # l = [k for k in imow[i] if port_number(k) == po]
    # if not l:
    #     imow[f"o{pi}@{mi}"]
    # else:
    #     k = l[0]
    #     mn = max(mode_number(k), mi)
    #     if mn != mode_number(k):
    #         imow[f"o{pi}@{mn}"] = imow[k]
    #         del imow[k]

    # if po not in imow[pi]:
    #     imow[pi]["o"][po] = mo
    # else:
    #     imow[pi]["o"][po] = max(imow[pi]["o"][po], mo)
